# -*- coding: utf-8 -*-
"""gbtestapp setup script."""
from setuptools import setup

if __name__ == "__main__":

    setup(
      name="gbtestapp",
      description="Not a real package, just for testing GravityBee.",
      package_dir={'': 'src'}
    )